#ifndef XVECT_H_INCLUDED
#	define XVECT_H_INCLUDED


/* 3D Vector arithmetic C library
 *
 * 1991-2002, Martin Strubel // onk@section5.ch
 *
 * $Id: vect.h,v 1.10 2003-06-15 14:55:04 strubi Exp $
 *
 * CONFIGURATION
 *
 *  DOUBLE_PRECISION : use double arithmetic
 *  OPTIMIZE_INLINE  : optimize with inline functions
 *  
 */

/* ------------------------------------------------------------------------------ */

#include <string.h>

/* Very small tolerance value */
#define EPSILON 0.000001

#ifdef OPTIMIZE_INLINE
#	define VEC_INLINE  inline
#else
#	define VEC_INLINE 
#endif 

/* ------------------------------------------------------------------------------ */
/* Type definitions */

#ifdef __linux__
#	include <values.h>
#else
#	if defined(__MINGW32__) || defined (WIN32)
#		define MAXFLOAT HUGE_VAL
#		define MAXDOUBLE HUGE_VAL
#	endif
#	include <math.h>
#endif

#ifdef DOUBLE_PRECISION
typedef double vfloat;
#	define VECT_MAXFLOAT MAXDOUBLE
#else
typedef float vfloat;
#	define VECT_MAXFLOAT MAXFLOAT
#endif

/** \addtogroup vect Vector math routines
 *
 * This documentation is yet incomplete.
 *
 * \todo complete vect library docs
 *
 */
/** @{ */

/** Vectors: vfloat triples */

typedef vfloat Vector3[3];
typedef vfloat *pVector3;

/** 4D vectors: needed for 3D perspective projections */
typedef vfloat Vector4[4];
typedef vfloat *pVector4;

/* Plane: a vfloat 4-tuple of coefficients A, B, C, D
 * 
 * where the Plane equation is:
 *
 * A*x + B*y + C*z = D     or     (A, B, C) dot (x, y, z) = D
 */

typedef vfloat Plane3[4];
typedef vfloat *pPlane3;

/** Matrices are stored in the column order:

\verbatim

A(x,y) = [ a[0][0] ...  a[n][0] ]
         [ .                    ]
         [ .                    ]
         [ .                    ]
         [ a[0][m] ...  a[n][m] ]

\endverbatim

a[0] is therefore regarded as a column vector which is a contiguous
chunk in memory. 
*/

typedef vfloat Matrix3[3][3];
typedef vfloat pMatrix3[][3];
typedef vfloat Matrix4[4][4];
typedef vfloat pMatrix4[][4];

/** Quaternions

 * Quaternions basically are a representation of a rotation around any arbitrary
 * axis in space. They are stored as a 4-tuple (w, x, y, z)
 *
 * See quat.c for more details.
 *
 */

typedef vfloat Quat[4];
typedef vfloat *pQuat;

/* ------------------------------------------------------------------------------ */
/* MACROS */

/** The determinant of the 2x2 matrix which results if row a and column b
   is blended out : */  

#define MAT2_DET(a1, b1, a2, b2) \
	m[a1][b1] * m[a2][b2] - m[a2][b1] * m[a1][b2] 

#define copy3(to, from) \
	memcpy(to, from, 3 * sizeof(vfloat))

#ifdef __cplusplus
extern "C" {
#endif

/* ------------------------------------------------------------------------------ */
/* GLOBALS */

extern vfloat g_epsilon;

/* ------------------------------------------------------------------------------ */
/* PROTOS  */
/* VECTOR FUNCTIONS */

/** Adds "y" to "x" */
VEC_INLINE
void add3(Vector3 x, const Vector3 y);

/** subtracts "y" from "x" */
VEC_INLINE
void sub3(Vector3 x, const Vector3 y);

/** x = -x  */
VEC_INLINE
void neg3(Vector3 x);

/** returns euclidian abs() (length) of vector "x" */
VEC_INLINE
vfloat len3(const Vector3 x);

/** Returns scalar product of "x" and "y" */
VEC_INLINE
vfloat dot3(const Vector3 x, const Vector3 y);

/** scales Vector "x" by scalar "s" */
VEC_INLINE
void scale3(Vector3 x, vfloat s);

/** normalizes vector "x" */
VEC_INLINE
vfloat norm3(Vector3 x);

/** stores cross product of vector x and y in out */
VEC_INLINE
void cross3(Vector3 out, const Vector3 x, const Vector3 y);





/* ------------------------------------------------------------------------------ 
 * MATRIX FUNCTIONS 
 */

/** print matrix (debugging) */
void printmat(pMatrix3 m);

/** init matrix "out" with unity matrix */
void unity3(pMatrix3 out);

/** copy matrix "src" to "dest" */
void copymat3(pMatrix3 dest, pMatrix3 src);

/** transpose matrix "m" */
void transmat(pMatrix3 m);

/** multiply vector "x" with matrix "m" and store result in "out" */
void matmulvec3(pVector3 out, pMatrix3 m, pVector3 x);

/** multiply matrix "m" with row vector "v" and store result in "out" */
void vecmulmat3(pVector3 out, pVector3 v, pMatrix3 m);

/** multiply matrix "a" with matrix "b" and store result in "out" */
void matmulmat3(pMatrix3 out, pMatrix3 a, pMatrix3 b);

/** scale matrix "m" by float value "s" */
void scalemat(pMatrix3 m, vfloat s);

/** This is a FAST inverse function. The matrix m is inverted by an algorithm
 * using the wonderful properties of the matrix' determinant (which is calculated
 * at the same time). 
 * NOTE! The inversion does not regard the scaling of the matrix, so to get 
 * the correct inverse, all matrix elements have to be divided through the
 * returned determinant (which should be checked for abs(det) < EPSILON, unless
 * it's sure that it is a regular matrix.
 *
 */
vfloat invmatdet3(pMatrix3 out, pMatrix3 m);

// simple conversions:

/** Build matrix "m" from a vector "z" and an up vector "y" */
void axistomat(pMatrix3 m, pVector3 y, pVector3 z);

/** Build rotation matrix "m" from euler rotation vector "rot" */
void rotmatrix(pMatrix3 m, pVector3 rot);

/* ------------------------------------------------------------------------------ 
 * 4x4 MATRIX ROUTINES                                                            
 */

void vec4tovec3(pVector3 out, const pVector4 in);
void vec3tovec4(pVector4 out, const pVector4 in);
void unity4(pMatrix4 out);
void mat4transpose(pMatrix4 mat);
void printmat4(const pMatrix4 m);
void copymat4(pMatrix4 dest, pMatrix4 src);
void mat3tomat4(pMatrix4 out, pMatrix3 in);
void mat4mulvec3(pVector3 out, const pMatrix4 m, const pVector3 x);
void mat4mulvec3_t(pVector3 out, const pMatrix4 m, const pVector3 x);
void mat4mulmat4(pMatrix4 out, const pMatrix4 a, const pMatrix4 b);
void mat4mulvec4_t(pVector4 out, const pMatrix4 mat, const pVector4 in);
void mat4mulvec4(pVector4 out, const pMatrix4 mat, const pVector4 in);
void mat4smartinv(pMatrix4 out, pMatrix4 m);

/* ------------------------------------------------------------------------------ 
 * ERROR HANDLING
 */

void vfatal(char *msg);

/* ------------------------------------------------------------------------------ 
 * Auxiliary vector routines
 */

void project(Vector3 v, const Vector3 ray);
void plane_fromface(pPlane3 plane, pVector3 v1, pVector3 v2, pVector3 v3);
int ray_plane_intersect(vfloat *where, const Vector3 ray,
		const Vector3 orig, const Plane3 plane);
int is_inside_triangle(const Vector3 point, const Vector3 v1,
		const Vector3 v2, const Vector3 v3);
int classify_halfspace(pPlane3 plane, pVector3 v, vfloat epsilon);

/** @} */

#ifdef __cplusplus
} // extern "C"
#endif
#endif // include once
